// Copyright 2015, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Author: Zhiwei Zhang <zhiwei.zhang@saturn.uni-freiburg.de>

#ifndef ZHIWEI_ZHANG_SOURCE_OTHER_GOLOMBENCODINGVECTOR_H_
#define ZHIWEI_ZHANG_SOURCE_OTHER_GOLOMBENCODINGVECTOR_H_

#include <vector>
#include "./EliasGammaEncodingVector.h"

using std::vector;

class GolombEncodingVector {
 public:

  // Encodes the gaps in @list via Golomb encoding using the mod parameter @M.
  template<typename Numeric>
  static void encode(const vector<Numeric>& plaintext, const int& M,
                       Bitvector& encoded, const bool& show = false) {
    Timer t;
    t.start();
    encoded.reserve(plaintext.size() * 10000);
    int nofRemainderBits = ceil(log2(M)) - 1;
    for (size_t i = 0; i < plaintext.size(); ++i) {
      /* since I use a gao list as input, so we don't need to calculate the gap.
       * if original list is given as input, just remove the comment mark and
       * add coment mark at line 37 and 38.
       *
       * Numeric gap = (i==0?plaintext[0]:plaintext[i]-plaintext[i-1]);
       * int q = gap / M;
       * Numeric r = gap - q * M;
       *
       */

      // Note: when M is a power of 2, this is not necessary.
      // r would be the log(M) least significant bits of gap.
      int q = plaintext[i] / M;
      Numeric r = plaintext[i] - q * M;
      for (int j = 0; j < q; ++j) encoded.push_back(0);
      encoded.push_back(1);
      // Write the remainder.
      // Find out the number of bits
      for (int j = nofRemainderBits; j >= 0; --j) {
        // Get the value of the bit in r that is nofBits - j from the right.
        encoded.push_back((r & (1 << j)) != 0);
      }
    }
    t.stop();
    if (show) {
      std::cout << plaintext.size() << "\t" << encoded.size()/8 << "\t"
      << t.getUSecs() << "\t";
    }
  }

  // Assumes @bits contains the gaps of a Doclist encoded with golomb M=@M,
  // Restores the Doclist.
  template<typename Numeric>
  static void decode(const Bitvector& bits, const int& M,
              vector<Numeric>& decoded, const bool& show = false) {
    Timer t;
    t.start();
    decoded.reserve(bits.size());
    /* since I use a gao list as input of compress, so we just need
     * to calculate the gap.
     * if original list is given as input of compress, just remove
     * the comment mark and add coment mark at line 74.
     *
     * Numeric lastId = 0;
     *
     */
    int nofRemainderBits = ceil(log2(M)) - 1;
    size_t i = 0;
    while (i < bits.size()) {
      Numeric lastId = 0;
      // Count the zeros.
      size_t q = i;
      while (bits[i] == 0) { ++i; }
      q = i - q;
      // At bits[i] there is the separator. Skip it.
      ++i;
      // Read the remainder at the next bits.
      Numeric r = 0;
      for (int j = nofRemainderBits; j >= 0; --j) {
        if (bits[i++] == 1) r += (1 << j);
      }
      // Add to the Doclist
      lastId += q * M + r;
      decoded.push_back(lastId);
    }
    t.stop();
    if (show) std::cout << t.getUSecs() << std::endl;
  }
};

#endif  // ZHIWEI_ZHANG_SOURCE_OTHER_GOLOMBENCODINGVECTOR_H_
