// Copyright 2015, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Author: Björn Buchhold <buchholb> and Zhiwei Zhang <zhiwei.zhang@saturn.uni-freiburg.de>

#pragma once
#include <stdint.h>
#include <assert.h>
#include <algorithm>
#include <cmath>
#include "../Timer.h"

/*
 * In this file all the variants of Simple8b algorithm contain
 * no more the object methods and constructors, deconstructor.
 * Each variant of Simple8b algorithm contains only the old encode
 * and new decode functions.
 */
namespace ad_utility_new {

  // ! Selector mask,
  // ! see: Anh & Moffat: "Index compression using 64-bit words."
  static const uint64_t SIMPLE8B_SELECTOR_MASK = 0x000000000000000F;

  // ! Selectors,
  // ! see: Anh & Moffat: "Index compression using 64-bit words."
  static const struct {
    unsigned char _itemWidth;
    unsigned char _groupSize;
    unsigned char _wastedBits;
    uint64_t _mask;
  } SIMPLE8B_SELECTORS[16] = {
    {0, 240, 60, 0x0000000000000000},  // selector 0
    {0, 120, 60, 0x0000000000000000},  // selector 1
    {1,  60,  0, 0x0000000000000001},  // selector 2
    {2,  30,  0, 0x0000000000000003},  // selector 3
    {3,  20,  0, 0x0000000000000007},  // selector 4
    {4,  15,  0, 0x000000000000000F},  // selector 5
    {5,  12,  0, 0x000000000000001F},  // selector 6
    {6,  10,  0, 0x000000000000003F},  // selector 7
    {7,   8,  4, 0x000000000000007F},  // selector 8
    {8,   7,  4, 0x00000000000000FF},  // selector 9
    {10,  6,  0, 0x00000000000003FF},  // selector 10
    {12,  5,  0, 0x0000000000000FFF},  // selector 11
    {15,  4,  0, 0x0000000000007FFF},  // selector 12
    {20,  3,  0, 0x00000000000FFFFF},  // selector 13
    {30,  2,  0, 0x000000003FFFFFFF},  // selector 14
    {60,  1,  0, 0x0FFFFFFFFFFFFFFF},  // selector 15
  };

  // This class uses the old encode function and the variant 1 of decode function.
  class Simple8bCodeNew1 {
  public:
    
    //! Encodes a list of elements that can be interpreted as numeric values
    //! using the Simple8b compression scheme.
    //! Returns the number of bytes in the encoded array, will always be a
    //! multiple of 8.
    //! Requires encoded to be preallocated with sufficient space.
    template<typename Numeric>
    static void encode(Numeric* plaintext, size_t nofElements,
                       uint64_t* encoded, const bool& show = false) {
      Timer t;
      t.start();
      size_t nofElementsEncoded = 0;
      size_t nofCodeWordsDone = 0;
      while (nofElementsEncoded < nofElements) {
        size_t itemsLeft = nofElements - nofElementsEncoded;
        // Count the number of consecutive 0's and decide if
        // selectors 0 or 1 can be used
        bool selector0 = true;
        bool selector1 = false;
        for (size_t i = 0; i < std::min<size_t>(240, itemsLeft); ++i) {
          if (plaintext[nofElementsEncoded + i] != 0) {
            if (i > 120) {
              selector0 = false;
              selector1 = true;
              break;
            } else {
              // Not possible to use selector 0 or 1
              selector0 = false;
              break;
            }
          }
        }
        if (selector0) {
          // Use selector 0 to compress 240 consecutive 0's.
          encoded[nofCodeWordsDone] = 0;
          nofElementsEncoded += 240;
          nofCodeWordsDone++;
          continue;
        }
        if (selector1) {
          // Use selector 1 to compress 120 consecutive 0's.
          encoded[nofCodeWordsDone] = 1;
          nofElementsEncoded += 120;
          nofCodeWordsDone++;
          continue;
        }
        // Try selectors for as many items per codeword as possible,
        // take the next selector whenever it is not possible.
        for (unsigned char selector = 2; selector < 16; ++selector) {
          uint64_t codeword = selector;
          size_t nofItemsInWord = 0;
          while (nofItemsInWord < std::min<size_t>(itemsLeft,
                           SIMPLE8B_SELECTORS[selector]._groupSize)) {
            // Check that the max value (60 bit) is not exceeded.
            assert(plaintext[nofElementsEncoded + nofItemsInWord]
                   <= 0x0FFFFFFFFFFFFFFF);
            // If an element is too large, break the loop.
            // Later we recognize that not enough elements
            // were written for the specific selector, and hence try the next
            // selector.
            if (plaintext[nofElementsEncoded + nofItemsInWord]
                > SIMPLE8B_SELECTORS[selector]._mask) {
              break;
            }
            codeword |= (static_cast<uint64_t>(
                         plaintext[nofElementsEncoded + nofItemsInWord])
                         << (4 +  // Selector bits.
                         nofItemsInWord * SIMPLE8B_SELECTORS[selector]._itemWidth));
            ++nofItemsInWord;
          }
          // Check if enough elements have been written to fit the selector
          // or if the loop was left earlier.
          if (nofItemsInWord == SIMPLE8B_SELECTORS[selector]._groupSize ||
              nofItemsInWord == itemsLeft) {
            encoded[nofCodeWordsDone] = codeword;
            nofElementsEncoded += nofItemsInWord;
            ++nofCodeWordsDone;
            break;
          }
        }
      }
      t.stop();
      if (show) {
        std::cout << nofElements << "\t"
        << sizeof(uint64_t) * nofCodeWordsDone << "\t"
        << t.getUSecs() << "\t";
      }
    }

    //! Decodes a list of elements using the Simple8b compression scheme.
    //! Requires decoded to be preallocated with sufficient space,
    //! i.e. sizeof(Numeric) * (nofElements + 239).
    //! The overhead is included so that no check for noundaries
    //! is necessary inside the decoding of a single codeword.
    // This is the variant 1 of the decode function.
    template<typename Numeric>
    static void decode(uint64_t* encoded, size_t nofElements,
                       Numeric* decoded, const bool& show = false) {
      Timer t;
      t.start();
      size_t nofElementsDone(0), nofCodeWordsDone(0);
      // Loop over full 64bit codewords
      while (nofElementsDone < nofElements) {
        unsigned char selector = encoded[nofCodeWordsDone]
        & SIMPLE8B_SELECTOR_MASK;
        // Decide whether the element in the encoded list contains only zero or not.
        if (encoded[nofCodeWordsDone] > selector) {
          uint64_t word = encoded[nofCodeWordsDone] >> 4;
          for (size_t i = 0; i < SIMPLE8B_SELECTORS[selector]._groupSize;
               ++i) {
            decoded[nofElementsDone++] = word
            & SIMPLE8B_SELECTORS[selector]._mask;
            word = word >> SIMPLE8B_SELECTORS[selector]._itemWidth;
          }
        } else {
          for (size_t i = 0;
               i < SIMPLE8B_SELECTORS[selector]._groupSize; ++i) {
            decoded[nofElementsDone++] = 0;
          }
        }
        ++nofCodeWordsDone;
      }
      t.stop();
      if (show) std::cout << t.getUSecs() << std::endl;
    }
  };

  // This class uses the old encode function and the variant 2 of decode function.
  class Simple8bCodeNew2 {
  public:
    
    //! Encodes a list of elements that can be interpreted as numeric values
    //! using the Simple8b compression scheme.
    //! Returns the number of bytes in the encoded array, will always be a
    //! multiple of 8.
    //! Requires encoded to be preallocated with sufficient space.
    template<typename Numeric>
    static void encode(Numeric* plaintext, size_t nofElements,
                       uint64_t* encoded, const bool& show = false) {
      Timer t;
      t.start();
      size_t nofElementsEncoded = 0;
      size_t nofCodeWordsDone = 0;
      while (nofElementsEncoded < nofElements) {
        size_t itemsLeft = nofElements - nofElementsEncoded;
        // Count the number of consecutive 0's and decide if
        // selectors 0 or 1 can be used
        bool selector0 = true;
        bool selector1 = false;
        for (size_t i = 0; i < std::min<size_t>(240, itemsLeft); ++i) {
          if (plaintext[nofElementsEncoded + i] != 0) {
            if (i > 120) {
              selector0 = false;
              selector1 = true;
              break;
            } else {
              // Not possible to use selector 0 or 1
              selector0 = false;
              break;
            }
          }
        }
        if (selector0) {
          // Use selector 0 to compress 240 consecutive 0's.
          encoded[nofCodeWordsDone] = 0;
          nofElementsEncoded += 240;
          nofCodeWordsDone++;
          continue;
        }
        if (selector1) {
          // Use selector 1 to compress 120 consecutive 0's.
          encoded[nofCodeWordsDone] = 1;
          nofElementsEncoded += 120;
          nofCodeWordsDone++;
          continue;
        }
        // Try selectors for as many items per codeword as possible,
        // take the next selector whenever it is not possible.
        for (unsigned char selector = 2; selector < 16; ++selector) {
          uint64_t codeword = selector;
          size_t nofItemsInWord = 0;
          while (nofItemsInWord
                 < std::min<size_t>(itemsLeft, SIMPLE8B_SELECTORS[selector]._groupSize)) {
            // Check that the max value (60 bit) is not exceeded.
            assert(plaintext[nofElementsEncoded + nofItemsInWord]
                   <= 0x0FFFFFFFFFFFFFFF);
            // If an element is too large, break the loop.
            // Later we recognize that not enough elements
            // were written for the specific selector, and hence try the next
            // selector.
            if (plaintext[nofElementsEncoded + nofItemsInWord]
                > SIMPLE8B_SELECTORS[selector]._mask) {
              break;
            }
            codeword |= (static_cast<uint64_t>(
                                               plaintext[nofElementsEncoded + nofItemsInWord])
                         << (4 +  // Selector bits.
                             nofItemsInWord * SIMPLE8B_SELECTORS[selector]._itemWidth));
            ++nofItemsInWord;
          }
          // Check if enough elements have been written to fit the selector
          // or if the loop was left earlier.
          if (nofItemsInWord == SIMPLE8B_SELECTORS[selector]._groupSize ||
              nofItemsInWord == itemsLeft) {
            encoded[nofCodeWordsDone] = codeword;
            nofElementsEncoded += nofItemsInWord;
            ++nofCodeWordsDone;
            break;
          }
        }
      }
      t.stop();
      if (show) {
        std::cout << nofElements << "\t"
        << sizeof(uint64_t) * nofCodeWordsDone << "\t"
        << t.getUSecs() << "\t";
      }
    }

    //! Decodes a list of elements using the Simple8b compression scheme.
    //! Requires decoded to be preallocated with sufficient space,
    //! i.e. sizeof(Numeric) * (nofElements + 239).
    //! The overhead is included so that no check for noundaries
    //! is necessary inside the decoding of a single codeword.
    // This is the variant 2 of the decode function.
    template<typename Numeric>
    static void decode(uint64_t* encoded, size_t nofElements,
                       Numeric* decoded, const bool& show = false) {
      Timer t;
      t.start();
      unsigned char selector(encoded[0] & SIMPLE8B_SELECTOR_MASK), i(0);
      // Loop over full 64bit codewords
      for (size_t nofElementsDone(0), nofCodeWordsDone(0);
           nofElementsDone < nofElements; ++nofCodeWordsDone,
           selector = encoded[nofCodeWordsDone] & SIMPLE8B_SELECTOR_MASK, i = 0) {
        // if the element in the encoded list contains only zeros
        // we do the decode there, otherwise we do it in the next for-loop
        for (; encoded[nofCodeWordsDone] == selector
             && i < SIMPLE8B_SELECTORS[selector]._groupSize;
             decoded[nofElementsDone++] = 0, ++i) { }

        for (size_t word(encoded[nofCodeWordsDone] >> 4);
             encoded[nofCodeWordsDone] != selector
             && i < SIMPLE8B_SELECTORS[selector]._groupSize;
             decoded[nofElementsDone++] = word & SIMPLE8B_SELECTORS[selector]._mask,
             word >>= SIMPLE8B_SELECTORS[selector]._itemWidth, ++i) { }
      }
      t.stop();
      if (show) std::cout << t.getUSecs() << std::endl;
    }
  };
}
